api_token = None
base = None
query_route = None
feed_route = None
source_route = None
upload_route = None
validate_route = None
file_size_limit = 250000000
upload_size_limit = 20000000
chunk_size = 200000
logger = None
col_ref = None
counties_ref = None
meta_base = None
